/*
 * $Id: corpadm.c,v 1.5 1996/02/26 22:52:40 scotty Rel $
 *
 * $Log: corpadm.c,v $
 * Revision 1.5  1996/02/26 22:52:40  scotty
 * Added support for braindead (Ultrix) syslog()s
 *
 * Revision 1.4  1996/02/23 21:51:50  scotty
 * Changed char *szDir to char rgchDir[2] in Users().
 * gcc (2.5.7 on skeat) didn't like what I was doing with it.
 *
 * Revision 1.3  1996/02/22 11:37:06  scotty
 * Now includes limits.h; OSF/1 was picking it up anyway but SunOS didn't.
 *
 * Revision 1.2  1996/02/15 02:12:17  scotty
 * Now sets the umask to ACC_UMASK (defined in Makefile).
 *
 * Revision 1.1  1996/02/15 01:27:36  scotty
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <syslog.h>
#include <limits.h>
#include "corpdefs.h"
#include "corpuser.h"
#include "corpfuns.h"


static char rcsid[] = "$Id: corpadm.c,v 1.5 1996/02/26 22:52:40 scotty Rel $";

#define NARGMAX   16
#define BUFLEN    4096
/*
 * for Users() and DisplayUsers()
 */
#define LEN       8
#define COLS      6
#define ROWS     10

static char    names[ROWS][COLS][LEN + 1];

static char   *szProgram;
static char   *rgszArgs[NARGMAX];
static int     idxArgs;

char *crypt(const char *, const char *);
/* void  CLine(char *); */
BOOL  CLine(char *, int *); 

#if defined(__ultrix__)
/*
 * Ultrix has no strdup(), so there's one in corpuser.c
 * which has to be declared here
 */
char *strdup(char *);
#endif

char  rgchSalt[] =
         "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/.";

int   List(void);
int   Users(void);
int   Add(void);
int   Delete(void);
int   Change(void);
int   Password(void);
int   Quit(void);
int   Help(void);
void  Usage(void);

void corperr(char *s, ...)
{
   return;
}



void  GetCommand(void);
char *GetString(char *);
void  Parse(void);
int   DisplayUsers(int);
typedef struct cmd {
   char  *szName;
   int   (*fnExec)(void);
   struct {
      char *szName;
      char *szDesc;
   } rgArgs[NARGMAX];
   char  *szDesc;
} CMD, *CMDP;

CMD rgcmdTable[] = {
   {
      "list",
      List,
      {
         "user",           "the user to be listed",
         0,                0
      },
      "List a user's details"
   },

   {
      "users",
      Users,
      {
         0,                0
      },
      "List all usernames"
   },

   {
      "add",
      Add,
      {
         "user",           "the user to be created",
         "realname",       "the user's real name",
         "organization",   "organization or department user belongs to",
         "contact",        "contact information, usually email address",
         "expiry",         "date when user's access expires",
         "password",       "initial password",
         0,                0
      },
      "Create a new user"
   },

   {
      "delete",            Delete,
      {
         "user",           "the user to be deleted",
         0,                0
      },
      "Delete user"
   },

   {
      "change",            Change,
      {
         "user",           "the user whose details will be changed",
         0,                0
      },
      "Interactively change a users details"
   },

   {
      "password",          Password,
      {
         "user",           "the user whose password will be changed",
         "pass",           "the new password",
         0,                0
      },
      "Change a users password"
   },

   {
      "help",              Help,
      {
         "command",        "the command about which help is required",
         0,                0
      },
      "Provide help on commands"
   },

   {
      "quit",              Quit,
      {
         0,                 0
      },
      "Quit the admin program"
   },

   {
      0,                   0,
      {
         0,                0
      },
      0
   }
};


main(int argc, char **argv)
{
   char          *szParam = PARAM;
   extern char   *optarg;
   extern int     optind;
   int            opt;
   int            arg;
   int            nerrs;

   if ((szProgram = strrchr(*argv, '/')) != 0)
      szProgram++;
   else
      szProgram = *argv;

   umask(ACC_UMASK);
#if defined(LOG_MAIL)
   openlog(szProgram, LOG_PID|LOG_NDELAY, LOG_FACILITY);
#else
   openlog(szProgram, LOG_PID);
#endif

   while ((opt = getopt(argc, argv, "p:")) != EOF) {
      switch (opt) {
       case 'p':
         szParam = optarg;
         break;

       default:
         Usage();
         exit(1);
      }
   }

   CLine(szParam, &nerrs);

   for (arg = 0; (optind + arg) < argc && arg < NARGMAX; arg++)
      rgszArgs[arg] = argv[optind + arg];

   if (arg == NARGMAX) {
      fprintf(stderr, "%s: Too many arguments\n", szProgram);
      exit(1);
   }

   for (; arg < NARGMAX; rgszArgs[arg++] = 0)
      ;

   Parse();

   exit(0);
}

void Usage(void)
{
   fprintf(stderr, "Usage: %s [-p prm_file] [command]\n",
           szProgram);
}

CMD *pcmdFindSz(char *szCmd)
{
   int   matches;
   CMD   *pcmdCurr;
   CMD   *pcmdSelect;
   
   for (matches = 0, pcmdCurr = rgcmdTable; pcmdCurr->szName; pcmdCurr++)
      if (!strncasecmp(szCmd, pcmdCurr->szName, strlen(szCmd)))
         if (++matches == 1)
            pcmdSelect = pcmdCurr;

   if (matches == 0) {
      fprintf(stderr, "%s: Unknown command '%s'\n",
              szProgram, szCmd);
      pcmdSelect = 0;
   } else if (matches > 1) {
      fprintf(stderr, "%s: Ambiguous command '%s'\n",
              szProgram, szCmd);
      pcmdSelect = 0;
   }

   return pcmdSelect;
}


void Parse(void)
{
   int   done = 0;
   int   once = 0;
   int   i;
   CMD   *pcmdSelect;

   if (rgszArgs[0])
      once = 1;

   do {
      if (!once) {
         rgszArgs[0] = 0;
         idxArgs = 0;
         GetCommand();
         if (rgszArgs[0] == 0)
            continue;
      }

      if (pcmdSelect = pcmdFindSz(rgszArgs[0])) {
         idxArgs = 1;
         done = pcmdSelect->fnExec();
      }
   } while (!done && !once);

}

void GetCommand(void)
{
   char  rgchBuf[BUFLEN];
   char *pchBuf;
   char *pchWord;
   int   idxCurr;

   fprintf(stdout, "%s> ", szProgram);
   fflush(stdout);
   fgets(rgchBuf, BUFLEN, stdin);

   /*
    * rgszArgs[] is only made up from malloc'd strings if the
    * program was called with no args, so it's safe to always
    * attempt to free the arguments.
    */
   for (idxCurr = 0; idxCurr < NARGMAX; rgszArgs[idxCurr++] = 0)
      if (rgszArgs[idxCurr] != 0) {
         free(rgszArgs[idxCurr]);
         rgszArgs[idxCurr] = 0;
      }

   for (pchBuf = rgchBuf, idxCurr = 0; idxCurr < NARGMAX; idxCurr++) {
      for (; *pchBuf && isspace(*pchBuf); pchBuf++)
         ;

      if (*pchBuf == '\0')
         break;

      for (pchWord = pchBuf; *pchBuf && !isspace(*pchBuf); pchBuf++)
         ;

      *pchBuf++ = '\0';

      if ((rgszArgs[idxCurr] = strdup(pchWord)) == 0) {
         fprintf(stderr, "%s: OUT OF MEMORY\n", szProgram);
         return;
      }
   }

   if (idxCurr == NARGMAX && *pchBuf) {
      fprintf(stderr, "%s: Too many arguments\n",
              szProgram);
   }

   idxArgs = 0;
}

char *GetString(char *szPrompt)
{
   char  rgchBuf[BUFLEN];
   char *pchBuf;
   char *pchWord;

   if (rgszArgs[idxArgs] != 0)
      return rgszArgs[idxArgs++];

   fprintf(stdout, "%s ", szPrompt);
   fflush(stdout);

   fgets(rgchBuf, BUFLEN, stdin);

   for (pchBuf = rgchBuf; *pchBuf && isspace(*pchBuf); pchBuf++)
      ;

   for (pchWord = pchBuf; *pchBuf && *pchBuf != '\n' && *pchBuf != '\r';
        pchBuf++)
      ;

   *pchBuf = '\0';

   if ((rgszArgs[idxArgs] = strdup(pchWord)) == 0) {
      fprintf(stderr, "%s: OUT OF MEMORY\n", szProgram);
      return 0;
   }

   return rgszArgs[idxArgs++];
}


int List(void)
{
   char *szUser;
   USR  *pusrInfo;
   char  rgchBuf[80];

   if ((szUser = GetString("User? ")) && *szUser) {
      if (pusrInfo = pusrRead(szUser)) {
         fprintf(stdout, "User                    : %s\n", pusrInfo->u_name);
         fprintf(stdout, "Real Name               : %s\n",
                 (pusrInfo->u_realname) ? pusrInfo->u_realname : "(none)");
         fprintf(stdout, "Contact                 : %s\n",
                 (pusrInfo->u_contact) ? pusrInfo->u_contact : "(none)");
         fprintf(stdout, "Organization            : %s\n",
                 (pusrInfo->u_organization) ? pusrInfo->u_organization
                                            : "(none)");
         if (pusrInfo->u_lastsucc != 0)
            strftime(rgchBuf, 80, "%I.%M%p, %A %d %B %Y",
                     localtime(&pusrInfo->u_lastsucc));
         else
            strcpy(rgchBuf, "(never)");
         fprintf(stdout, "Last Successful Login   : %s\n", rgchBuf);
         if (pusrInfo->u_lastunsucc != 0)
            strftime(rgchBuf, 80, "%I.%M%p, %A %d %B %Y",
                     localtime(&pusrInfo->u_lastunsucc));
         else
            strcpy(rgchBuf, "(never)");
         fprintf(stdout, "Last Unsuccessful Login : %s\n", rgchBuf);
         if (pusrInfo->u_expires != 0)
            strftime(rgchBuf, 80, "%A %d %B %Y",
                     localtime(&pusrInfo->u_expires));
         else
            strcpy(rgchBuf, "(never)");
         fprintf(stdout, "Account expires         : %s\n", rgchBuf);
      } else {
         fprintf(stdout, "No user %s\n", szUser);
      }
   } else if (szUser)
      Users();

   return 0;
}

int Add(void)
{
   USR         usrInfo;
   char       *szDate = "";
   char       *szPwd1 = "";
   char       *szPwd2 = "";
   char        salt[3];
   struct tm   tmExp;

   if (usrInfo.u_name = GetString("User? ")) {
      if (pusrRead(usrInfo.u_name)) {
         fprintf(stderr, "User %s already exists\n", usrInfo.u_name);
         return 0;
      }

      if ((usrInfo.u_realname = GetString("Real name? ")) == 0)
         return 0;

      if ((usrInfo.u_organization = GetString("Organization? ")) == 0)
         return 0;

      if ((usrInfo.u_contact = GetString("Contact Address? ")) == 0)
         return 0;

      do {
         if ((szDate = GetString("Expiration Date (dd/mm/yyyy)? ")) == 0)
            return 0;
      } while (*szDate &&
               sscanf(szDate, "%d/%d/%d",
                      &tmExp.tm_mday, &tmExp.tm_mon, &tmExp.tm_year) != 3);
      if (*szDate) {
         tmExp.tm_sec = tmExp.tm_min = 0;
         tmExp.tm_hour = 4;
         tmExp.tm_wday = tmExp.tm_yday = tmExp.tm_isdst = -1;
         tmExp.tm_mon--;
         tmExp.tm_year -= 1900;

         if ((usrInfo.u_expires = mktime(&tmExp)) == (time_t)-1) {
            fprintf(stderr, "Invalid date -- '%s'\n", szDate);
            return 0;
         }
      } else
         usrInfo.u_expires = (time_t)0;

      do {
         if ((szPwd1 = GetString("Password? ")) == 0)
            return 0;
      } while (*szPwd1 == '\0');

      salt[0] = rgchSalt[rand() % 64];
      salt[1] = rgchSalt[rand() % 64];
      salt[2] = '\0';

      usrInfo.u_pwd = crypt(szPwd1, salt);

      usrInfo.u_lastsucc = usrInfo.u_lastunsucc = 0;

      if (WritePUsr(&usrInfo)) {
         fprintf(stderr, "%s: Couldn't create user %s\n",
                 szProgram, usrInfo.u_name);
         perror("WritePUsr");
         return 0;
      } else {
         fprintf(stdout, "%s: Created user %s\n",
                 szProgram, usrInfo.u_name);
      }
   }
   return 0;
}

int Delete(void)
{
   char  path[_MAX_PATH];
   char *szName;

   if ((szName = GetString("User? ")) == 0)
      return 0;

   sprintf(path, "%s%c/%s", accpath, *szName, szName);

   if (unlink(path) < 0) {
      fprintf(stderr, "%s: Couldn't delete user %s\n",
              szProgram, szName);
      perror("unlink:");
   } else {
      fprintf(stdout, "%s: Deleted user %s\n",
              szProgram, szName);
   }
   return 0;
}


int Change(void)
{
   char       *szUser;
   USR        *pusrInfo;
   char        rgchBuf[200];
   char       *szNew;
   char       *szDate;
   struct tm   tmExp;

   if (szUser = GetString("User? ")) {
      if (pusrInfo = pusrRead(szUser)) {
         sprintf(rgchBuf, "Real name [%s]?\n> ",
                 pusrInfo->u_realname);
         if ((szNew = GetString(rgchBuf)) == 0)
            return 0;
         if (*szNew)
            pusrInfo->u_realname = szNew;

         sprintf(rgchBuf, "Organization [%s]?\n> ",
                 pusrInfo->u_organization);
         if ((szNew = GetString(rgchBuf)) == 0)
            return 0;
         if (*szNew)
            pusrInfo->u_organization = szNew;

         sprintf(rgchBuf, "Contact Address [%s]?\n> ",
                 pusrInfo->u_contact);
         if ((szNew = GetString(rgchBuf)) == 0)
            return 0;
         if (*szNew)
            pusrInfo->u_contact = szNew;

         if (pusrInfo->u_expires)
            strftime(rgchBuf, 80, "Expiration Date [%d/%m/%Y]?\n> ",
                     localtime(&pusrInfo->u_expires));
         else
            sprintf(rgchBuf, "Expiration Date [never]?\n> ");
         do {
            if ((szDate = GetString(rgchBuf)) == 0)
               return 0;
         } while (*szDate &&
                  sscanf(szDate, "%d/%d/%d",
                         &tmExp.tm_mday, &tmExp.tm_mon, &tmExp.tm_year) != 3);
         if (*szDate) {
            tmExp.tm_sec = tmExp.tm_min = 0;
            tmExp.tm_hour = 4;
            tmExp.tm_wday = tmExp.tm_yday = tmExp.tm_isdst = -1;
            tmExp.tm_mon--;
            tmExp.tm_year -= 1900;

            if ((pusrInfo->u_expires = mktime(&tmExp)) == (time_t)-1) {
               fprintf(stderr, "Invalid date -- '%s'\n", szDate);
               return 0;
            }
         } else
            pusrInfo->u_expires = (time_t)0;
         
         if (WritePUsr(pusrInfo)) {
            fprintf(stderr, "%s: Couldn't change user %s\n",
                    szProgram, pusrInfo->u_name);
            perror("WritePUsr");
            return 0;
         } else {
            fprintf(stdout, "%s: Changed user %s\n",
                    szProgram, pusrInfo->u_name);
         }
         
      } else {
         fprintf(stdout, "No user %s\n", szUser);
      }
   }
      

   return 0;
}

int Password(void)
{
   char *szUser;
   char *szPass;

   if ((szUser = GetString("User? ")) == 0 || !*szUser)
      return 0;

   if ((szPass = GetString("Password? ")) == 0)
      return 0;

   if (!*szPass)
      fprintf(stderr, "%s: Null passwords are not allowed\n",
              szProgram);
   else {
      if (SetPassword(szUser, (char *)-1, szPass) != 1)
         fprintf(stderr, "%s: Couldn't change password\n", szProgram);
      else
         fprintf(stdout, "%s: Password changed\n", szProgram);
   }

   return 0;
}

int Help(void)
{
   char *szCommand;
   CMD  *pcmdHelp;
   int   iargCurr;

   if ((szCommand = GetString("Command? ")) == 0)
       return 0;

   if (!*szCommand) {
      fprintf(stdout, "Commands are:\n\n");
      for (pcmdHelp = rgcmdTable; pcmdHelp->szName; pcmdHelp++)
         fprintf(stdout, "  %-20.20s - %s\n",
                 pcmdHelp->szName, pcmdHelp->szDesc);
      fprintf(stdout, "\n");
   } else {
      if (pcmdHelp = pcmdFindSz(szCommand)) {
         fprintf(stdout, "\nCOMMAND:\n  %-20.20s - %s\n\n",
                 pcmdHelp->szName, pcmdHelp->szDesc);
         if (pcmdHelp->rgArgs[0].szName) {
            fprintf(stdout, "ARGUMENTS:\n");
            for (iargCurr = 0; pcmdHelp->rgArgs[iargCurr].szName;
                 iargCurr++) {
               fprintf(stdout, "  %-20.20s - %s\n",
                       pcmdHelp->rgArgs[iargCurr].szName,
                       pcmdHelp->rgArgs[iargCurr].szDesc);
            }
         }
         fprintf(stdout, "\n");
      }
   }
   
   return 0;
}

int Users(void)
{
   DIR           *pdirCurr;
   struct dirent *pdeCurr;
   int            count = 0;
   char           rgchDir[2];
   int            done = 0;

   if (chdir(accpath) != 0) {
      fprintf(stderr, "%s: Can't cd to accounts directory %s\n",
              szProgram,accpath);
      return 0;
   }

   for (rgchDir[0] = 'a', rgchDir[1] = '\0';
        !done && rgchDir[0] <= 'z'; rgchDir[0]++) {
      if ((pdirCurr = opendir(rgchDir)) == 0) {
         /*
          * Try and create missing directories
          */
         fprintf(stderr, "%s: Trying to create accounts directory %s... ",
                 szProgram, rgchDir);
         if (mkdir(rgchDir, 0770) == 0)
            fprintf(stderr, "OK\n");
         else
            fprintf(stderr, "FAILED\n");
         continue;
      }

      while (!done && (pdeCurr = readdir(pdirCurr)) != 0) {
         if (*pdeCurr->d_name >= 'a' && *pdeCurr->d_name <= 'z') {
            strncpy(names[count / COLS][count % COLS], pdeCurr->d_name,
                    LEN);
            count++;
         }
         
         if (count == (ROWS * COLS)) {
            if (!DisplayUsers(count))
               done = 1;
            count = 0;
         }
      }
      closedir(pdirCurr);
   }

   if (count)
      DisplayUsers(count);

   return 0;
}


int DisplayUsers(int count)
{
   int   i;
   int   ch;

   for (i = 0; i < count; i++) {
      if ((i % COLS) == 0) {
         if (i)
            printf("\n");
         printf("    ");
      }
      printf("%-*.*s   ", LEN, LEN, names[i / COLS][i % COLS]);
   }
   if (count == (ROWS * COLS)) {
      if (i % COLS)
         printf("\n");
      printf("\nRETURN for more");
      ch = getchar();
      return (ch == '\n' || ch == '\r');
   } else {
      printf("\n");
      return 1;
   }
}


int Quit(void)
{
   fprintf(stdout, "Quitting...\n");
   return 1;
}

