#include "corpinc.h"                   

#define MAXLINE 256         
#define MAXK 16                                     

enum keys {KEY_ELT,KEY_ATT,KEY_POS,KEY_PUN,KEY_CHAR,KEY_VER,KEY_ENT,KEY_LABEL,KEY_SCOPE,
	KEY_OPTION,KEY_LC,KEY_LEX,KEY_RADIX,KEY_LEMMATA,
	KEY_WTAG,KEY_LTAG,KEY_MAX};

char* kw[KEY_MAX]={"elt","att","pos","pun","char","ver","ent",
	"label","scope","option","lc","lex","radix","lemmata","wtag",
	"ltag"};
                            
void ProcessDescLine(char*);                            
void ProcessElt(char*);
void ProcessAtt(char*);
void ProcessPos(char*);
void ProcessPun(char*);
void ProcessChar(char*);                                           
void ProcessEnt(char*);
void ProcessLabels(char*);
void ProcessScope(char*);
void ProcessOption(char*);
void ProcessLC(char*);
void ProcessLex(char*);
void ProcessRadix(char*);
void ProcessLemmata(char*);
void ProcessWTag(char*);
void ProcessLTag(char*);

int npos, nctab, nspec, nelts, natts, nukatts, nukents, nukpos, nlabs, ncase,
	nwt, nlt=0;
int hdv;                                
int radix=10;

struct _wtg wtags[MAXWTAGS];
struct _wtg ltags[MAXWTAGS];
struct _pos postab[MAXENTS];
struct cdef ctab[MAXCENTS];        
struct _ent specs[MAXSPECS];  
struct _eltab eltab[MAXELTS];            
struct _atttab atttab[MAXGATTS];
struct _ukatttab uktab[MAXGATTS];
struct _lc lctab[MAXLC];
enum _st lextab[256];
char uketab[MAXGENTS][MAXNAME+1];
char ukpos[MAXENTS][MAXNAME];

static char currelt[MAXNAME+1]="0";                                  

BOOL bNoPos=FALSE;      
BOOL bNameCase=FALSE;

char txt0[MAXNAME]="TXT0";

char defscope[MAXNAME];

#define NTYPES 9                        
#define MAXTLENGTH 16

#define MAXLAB 10

struct _labels labels[MAXLAB];

static char* atype[NTYPES]={"CDATA","CAT","NUMBER","NAME","REFID","ID","NULL",
				"MULTID","MULTIDREFS"};
static char descbuff[MAXGDESC];
static int id=0;

void BuildDefLexTab();
int GetCh(char*);
                            
BOOL ReadDescFile(char* file)
{
	FILE* fDesc; char buffer[MAXLINE];
	BuildDefLexTab();
	npos=0; nelts=0; natts=0;
	fDesc=fopen(file,"r");
	if (fDesc==NULL) return FALSE;             
	while (fgets(buffer,MAXLINE,fDesc)) 
		ProcessDescLine(buffer);
	fclose(fDesc);
	return TRUE;
}

void ProcessDescLine(char* line)
{                                
	char* s=line; char key[MAXK];                                                     
	int i;
	if (line[0]=='#') return; /* comment */
	while (*s && (*s != '\n')) s++;
	*s=0;
	while (*line && isspace(*line)) line++;
	if (*line==0) return;
	s=key;
	while (*line && !isspace(*line)) *(s++)=*(line++);
	*s=0;             
	while (isspace(*line)) line++;
	for (i=0;i<KEY_MAX;i++)
		if(!_stricmp(key,kw[i])) break;
	if (i==KEY_MAX) return; /* note other users of the file may include lines, so this is not an error */
	switch (i) {
	case KEY_ELT:
		ProcessElt(line);
		break;
	case KEY_ENT:
		ProcessEnt(line);
		break;
	case KEY_ATT:
		ProcessAtt(line);
		break;
	case KEY_POS:
		ProcessPos(line);
		break;                                   
	case KEY_PUN:
		ProcessPun(line);
		break;	
	case KEY_CHAR:
		ProcessChar(line);
		break;	                                                               
	case KEY_VER:
		hdv=atoi(line);
		break;	 
	case KEY_LABEL:
		ProcessLabels(line);
		break;
	case KEY_SCOPE:
		ProcessScope(line);
		break;
	case KEY_OPTION:
		ProcessOption(line);
		break;
	case KEY_LC:
		ProcessLC(line);
		break;
	case KEY_LEX:
		ProcessLex(line);
		break;
	case KEY_RADIX:
		ProcessRadix(line);
		break;
	case KEY_LEMMATA:
		ProcessLemmata(line);
		break;
	case KEY_WTAG:
		ProcessWTag(line);
		break;
	case KEY_LTAG:
		ProcessLTag(line);
		break;
	}
}

enum _opt {OPT_NOPOS,OPT_NAMECASE,OPT_MAX};

char* opt[OPT_MAX]={"noposindex","namecase"};

void ProcessOption(char* l)
{
	char option[16]; int i;
	sscanf(l," %s",option);
	for (i=0;i<OPT_MAX;i++)
		if (!_stricmp(option,opt[i])) break;
	if (i==OPT_MAX) return;
	switch (i) {
	case OPT_NOPOS:
		bNoPos=TRUE;
		break;
	case OPT_NAMECASE:
		bNameCase=TRUE;
		break;
	default:
		break;
	}
}

void ProcessScope(char* l)
{
	static BOOL bScope=FALSE;
//	sscanf(l,"%s",txt0);
	if (bScope) return;
	sscanf(l," %s",defscope);
	bScope=TRUE;
}

void ProcessLabels(char* l)
{
	sscanf(l," %[^/]/%s",labels[nlabs].element,labels[nlabs].attribute);
	nlabs++;
}

void ProcessPos(char* name)
{
	int i=0;
	while (*name && !isspace(*name))
		postab[npos].name[i++]=*(name++);   
	postab[npos].name[i]=0;	
	postab[npos].bPun=FALSE;	
	npos++;
}

void ProcessPun(char* name)
{
	int i=0;
	while (*name && !isspace(*name))
		postab[npos].name[i++]=*(name++);   
	postab[npos].name[i]=0;
	postab[npos].bPun=TRUE;	
	npos++;
}

int atoh(char* s)
{
	int k=0; int d;
	char c;
	while (c=*s++) {
		if (c < '0') break;
		else if (c <= '9') d=c-'0';
		else if (c < 'A') break;
		else if (c <= 'F') d=c-'A'+10;
		else if (c < 'a') break;
		else if (c <= 'f') d=c-'a'+10;
		else break;
		k=16*k+d;
	}
	return k;
}

void ProcessChar(char* name)
{
	int i=0;
	while (*name && !isspace(*name))
		ctab[nctab].cname[i++]=*(name++);     
	ctab[nctab].cname[i]=0;	
	while (isspace(*name)) name++;
	if (isdigit(*name))
		ctab[nctab].val=strtoul(name,NULL,radix);
	else 
		ctab[nctab].val=-1;
	nctab++;
}

void ProcessLC(char* name)
{
	char x[100],y[100];
	sscanf(name,"%s %s",x,y);
	lctab[ncase].u=GetCh(x);
	lctab[ncase].l=GetCh(y);
	ncase++;
}

void ProcessLex(char* name)
{
	enum _st st;
	char s[100];
	char c; int k;
	sscanf(name,"%s %c",s,&c);
	k=GetCh(s);
	if (k < 0 || k > 255) return;
	switch (c) {
	case 's': case 'S':
		st=ST_SPACE;
		break;
	case 'c': case 'C':
		st=ST_CHAR;
		break;
	}
	lextab[k]=st;
}

void ProcessEnt(char* name)
{
	int i=0;
	while (*name && !isspace(*name))
		specs[nspec].name[i++]=*(name++);     
	specs[nspec].name[i]=0;	   
	while (*name && isspace(*name)) name++;         
	i=0;
	while (*name && !isspace(*name))
		specs[nspec].disp[i++]=*(name++);     
	specs[nspec].disp[i]=0;	
	nspec++;
}                                                                            

void ProcessElt(char* line)
{   
	int i=0; char c; int t;
	while (*line && !isspace(*line))
		eltab[nelts].name[i++]=*(line++);       
	eltab[nelts].name[i]=0;	
	strcpy(currelt,eltab[nelts].name);
	while (isspace(*line)) line++;
	c=*line;
	line++;
	switch (c) {
	case 's':
		t=LG_SENT;
		break;
	case '0':
		t=LG_EMPTY;
		break;
	case 'e':		
		t=LG_ELEMENT;
		break;
	}                         
	eltab[nelts].eltype=t;
	eltab[nelts].st=ST_SPACE;
	eltab[nelts].bInternal=FALSE;
	while (isspace(*line)) line++;
	while (isalnum(*line)) {
		switch(*line) {
		case 't':
			eltab[nelts].st=ST_DEL;
			break;
		case 'i':
			eltab[nelts].bInternal=TRUE;
			break;
		default:
			break;
		}
		line++;
	}
	nelts++;
}	

void ProcessRadix(char* line)
{
	radix=atol(line);
}

void ProcessAtt(char* line)
{        
	char type[MAXTLENGTH];
	int i=0; int t;
	while (*line && !isspace(*line))
		atttab[natts].name[i++]=*(line++);  
	atttab[natts].name[i]=0;	  
	while (isspace(*line)) line++;
	strcpy(atttab[natts].element,currelt);
	atttab[natts].val=natts+4;
	i=0;
	while (*line && !isspace(*line))
		type[i++]=*(line++);  
	type[i]=0;	
	for (t=0;t<NTYPES;t++)
		if (!_stricmp(atype[t],type)) break;      
	if (t==NTYPES) corperr("Bad attribute type %s",type);	
	if (t==AT_MULTIDREF) t |= AT_PLURAL;
	atttab[natts].attv=t;
	while (isspace(*line)) line++;  
	atttab[natts].attdesc=descbuff+id;
	while (*line && !isspace(*line))
		descbuff[id++]=*(line++);
	descbuff[id++]=0;	
	natts++;
}

void BuildDefLexTab()
{
	int i;
	for (i=0;i<256;i++) {
		if (i=='_' || i=='\'') lextab[i]=ST_CHAR;
		else if (isspace(i) || ispunct(i)) lextab[i]=ST_SPACE;
		else lextab[i]=ST_CHAR;
		if (i>128) lextab[i]=ST_CHAR;
	}
	lextab[0]=ST_DEL;
}

int GetCh(char* s)
{
	int i;
	if (s[0]==0) return 0;
	if (s[0]=='&') {
		int k=strlen(s);
		if (s[k-1]==';') s[k-1]=0;
		for (i=0;i<nctab;i++) {
			if (!strcmp(s+1,ctab[i].cname)) break;
		}
		if (i < nctab) return ctab[i].val; else return 0;
	}
	return (unsigned char) s[0];
}

void ProcessLemmata(char* line)
{
	char scheme[100]; int i=0;
	while (*line && !isspace(*line))
		scheme[i++]=*(line++);
	scheme[i]=0;
	for (i=0;i<LemmaNSchemes();i++) 
		if (!strcmp(scheme,LemmaName(i))) break;
	if (i==LemmaNSchemes()) return;
	while (*line && isspace(*line)) line++;
	LemmaEnable(i,line);
}

void ProcessWTag(char* line)
{
	int i=0;
	while (*line && !isspace(*line))
		wtags[nwt].name[i++]=*(line++);
	while (*line && isspace(*line)) line++;
	i=0;
	while (*line && !isspace(*line))
		wtags[nwt].pos[i++]=*(line++);
	nwt++;
}
		
			
void ProcessLTag(char* line)
{
	int i=0;
	while (*line && !isspace(*line))
		ltags[nlt].name[i++]=*(line++);
	while (*line && isspace(*line)) line++;
	i=0;
	while (*line && !isspace(*line))
		ltags[nlt].pos[i++]=*(line++);
	nlt++;
}
		
			



		
	

	
						
	 		
		
	
	